/***************************************************************/
/*  ADDQUART : version of 30/8/99 - this program is part of the 
               PHYLOQUART package v1.4  .                        
    (vberry@lirmm.fr)                                         

    Please distribute with the full package                    */
/***************************************************************/

/* Impovements wrt previous versions:

   - nouveaux parametres : 
        -f : indique degre de resolution souhaite (naturel ou maximum)
        -t : demarre de l'arbre etoile ou de l'arbre produit par Qstar.
   - accepte des quadruplets ayant un poids. 
   - accepte des noms d'especes de plus de 10 cars (jusqu'a 30).
   - lit les quads (avec eventuellement poids) en 1 instruct : fgets
   - sauve arbre sous forme parenthesee (enracine arbitrairement a l'esp 1)
     dans "treefile"  (fonction Sauve_Arbre)
*/
/***************************************************************************/
/*  
 
   ALGORITHME inferant un arbre en resolvant progressivement les noeuds
   de degre > 3 presents dans l'arbre infere par la methode Q*
   (programme qstar).

   Note : on a la propriete que les differents noeuds irresolus de
   l'arbre initial concernent des ensembles de quadruplets disjoints.

   Specifs : 
   - si apres Qstar : on part du fichier "bipfile", et par l'algo
   lineaire de tree-popping on constitue une struct d'arbre en memoire
   representant ||*. 
   - si pas apres Qstar : on part de l'arbre etoile.

   Puis on resoud chq noeud de degre > 3  separement
   par un algorithme glouton en O(n^4) semblable a Addtree.

ATTENTION : on peut forcer l'algo a inferer un arbre totalement
            resolu. C'est ok si on veut optimiser le critere NMQS 
	    (Nombre Max de Quadruplets Satisfaits), mais ca
	    n'est pas toujours approprie si on cherche le meilleur
	    estimateur de la phylogenie des taxa etudies.

*/
/*-------------------------------------------------------------------------*/
/* PARAM = -a (valeur de "a" ds critere de construction : SAT-a.CONTREDITS) 
           -t starts from a trivial tree (-t1) or from the tree coded in 'bipfile' (-t0)
           -f forces the inference of a fully resolved tree (-f1)          */
/* INPUT = 

     fic "quartfile" :  ensemble Q des quadruplets d'entree

 OR
 
     fic "bipfile" : liste des bipartitions de l'arbre a completer.
         Si on desire partir de la topologie etoile (la moins
	 resolue), le fichier "bipfile" doit contenir la correspondance 
         numero-taxa et les bipartitions triviales (un taxon oppose a 
         tous les autres). 
     fic "quartfile.left" : liste des quadruplets a resoudre (donc
	 non-resolus par les bipartitions de "bipfile") pour completer
	 l'arbre (doit contenir la correspondance numero-taxa).	 
     fic "quartfile.res" : liste des quadruplets resolus par l'arbre
     que AddQuart complete (doit contenir la correspondance numero-taxa)    */

/* OUTPUT = 
    fic "bipfile" complete par les bips induites par l'algo AddQuart        */
/*--------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "res-score.c"

/*--------------------------------------------------*/
void Affiche_Arbre(noeud,pere)
tarbre *noeud,*pere;
{
int i,prem;
/* Produit un arbre ENRACINE depuis la structure de donnees */

/* affiche ses taxa */
for (i=0;noeud->taxa[i];i++) printf("%.2d:[n%d]",noeud->taxa[i],noeud->num);

/* recurse */
if (noeud->voisins[1]) /* a un autre voisin que son pere */
    {
     printf("(");prem=1;
     for (i=0;noeud->voisins[i];i++)
       {
	  if (noeud->voisins[i]!=pere)  
	    {if (!prem) printf(",");
	     else prem=0;
	     Affiche_Arbre(noeud->voisins[i],noeud);
	    }
	}
     printf("):[n%d]",noeud->num);
    }
}
/*-----------------------------------------*/
void Lit_quad_restant(star_tree,nbquads)
int star_tree,*nbquads;
/* SE : lit les r4-arbres de Q qui restent encore a resoudre */
/* SI : charge les quads depuis fichier dans MatQ */
{
FILE *fquad;
char chaine[100],nomfic[15];
int t1,t2,t3,t4,tmp,tmp2;
int nblus;
double score;
if (star_tree) strcpy(nomfic,"quartfile"); else strcpy(nomfic,"quartfile.left"); 
/* init l'ens de quads */
for (t1=1;t1<=Nb_taxa;t1++)
  for (t2=t1+1;t2<=Nb_taxa;t2++)
    for (t3=t1+1;t3<=Nb_taxa;t3++)
      for (t4=t3+1;t4<=Nb_taxa;t4++)
        MatQ[t1][t2][t3][t4]=0;
if (!(fquad=fopen(nomfic,"r"))) 
    Erreur("QUARTET FILE NOT FOUND",6);

/* Lit le nb de taxa */
fscanf(fquad,"%d",&Nb_taxa);
printf("%d taxa\n",Nb_taxa);
if (Nb_taxa>30) Erreur("Sorry, this program can't handle more than 30 taxa",0);
/* Lit la correspondance numero-espece */
for (tmp=1;tmp<=Nb_taxa;tmp++)
  fscanf(fquad,"%d %s\n",&tmp2,name[tmp]);
fscanf(fquad,"\n");

/* Lit les quadruplets */
while (fgets(chaine,100,fquad))
  {
   /* nouveau quadruplet */
   (*nbquads)++;
   /* extrait les 4 especes et le score si il est present */
   score = 0;
   /*printf("A DECODER = '%s'",chaine);*/
   nblus=sscanf(chaine,"%d %d || %d %d (%lf)",&t1,&t2,&t3,&t4,&score);
   /*printf("j'ai lu %d variables : %d %d || %d %d",nblus,t1,t2,t3,t4);*/
   /*if (nblus==5) printf(" score = %lf\n",score); else printf("\n");*/

   /* ordonne les taxas par ordre croissant */
   if (t1>t2) {tmp=t1;t1=t2;t2=tmp;}
   if (t3>t4) {tmp=t3;t3=t4;t4=tmp;}
   /* Enregistre le quad avec son score. 
      Un score de 1 est donne si pas de score specifique */
   if (t1<t3) {
	       if (score) MatQ[t1][t2][t3][t4]=score;
	       else MatQ[t1][t2][t3][t4]=1; }
   else {
         if (score)  MatQ[t3][t4][t1][t2]=score;
	 else MatQ[t3][t4][t1][t2]=1;}
  }
}
/*-------------------------------------------*/
void Cree_bip_triviales()
{
FILE *f;
int i,j,k;
printf("Creating trivial bipartitions\n");
f=fopen("bipfile","w");fprintf(f,"   %d\n",Nb_taxa);
for (i=1;i<=Nb_taxa;i++) fprintf(f,"%.2d %s\n",i,name[i]);
fprintf(f,"\n");
for (i=1;i<=Max_nb_noeuds;i++) for(k=0;k<Max_nb_taxa;k++)
  {bipartition[i][1][k]=0;bipartition[i][2][k]=0;}
for (i=1;i<=Nb_taxa;i++)
  {
  k=0;
  bipartition[i][1][0]=i;fprintf(f,"%.2d |",i);
  for (j=1;j<=Nb_taxa;j++)
    if (i!=j) {bipartition[i][2][k]=j;k++;fprintf(f," %.2d",j);}
  fprintf(f,"\n");
  }
Nb_Bips=Nb_taxa;fclose(f);
}
/*-------------------------------------------*/
void Lit_bipartitions(star_tree)
int star_tree; /* whether starts from the star (trivial) tree */ 
{
FILE *f;
char s[50];
int numbip=1,n=0,partie=1,nbobj=0;
int i,k,combien;
/* RAZ les bips */
for (i=1;i<=Max_nb_noeuds;i++) for(k=0;k<Max_nb_taxa;k++)
  {bipartition[i][1][k]=0;bipartition[i][2][k]=0;}
/* Start from trivial bips or read bipartitions inferred by qstar */
if ((star_tree) || (!(f=fopen("bipfile","r")))) 
  {
   Cree_bip_triviales();return;  /* demarre de l'arbre etoile */ 
  }
/* Lit le nb de taxa */
printf("Reading bipartitions of the initial tree from 'bipfile'\n");
fscanf(f,"%d",&combien);
if (combien!=Nb_taxa) Erreur("INCOHERENT NUMBER OF TAXA BETWEEN bipfile AND quartfile.left FILES!",-2);
/* Lit la correspondance numero-espece mais ne la garde pas, on la
   connait deja du fichier "quartfile.left" */
for (i=1;i<=Nb_taxa;i++)
  if (fscanf(f,"%d %s\n",&k,s)==-1) 
    Erreur("INCORRECT FORMAT in TAXA NAMES PART OF bipfile",7);
fscanf(f,"\n");

/* Lit les bipartitions */
while (fscanf(f,"%s",s)!=-1)
   {
     /*printf("%s ",s);*/nbobj++;
    if (!strcmp(s,"|")) {partie++;n=0;}
    else {bipartition[numbip][partie][n]=atoi(s);n++;}
    if (nbobj==Nb_taxa+1) /* fin de la bip courante */
	{/* lit le poids de la bipartition*/
	  /*fscanf(f,"%s",s); printf("le pds de la bip est %s\n",s);*/
	  n=0;nbobj=0;numbip++;partie=1;/*printf("\n");*/}
   }
Nb_Bips=numbip-1;fclose(f);
}
/*--------------------------------------------*/

/*_________________________________________________*/
/*_______________  algo de tree-popping ___________*/
/*_________________________________________________*/

void Profondeur(Reste,Atteindre,marques,ind_pile,Pile)
int *Reste,Atteindre[Max_nb_taxa+1],marques[Max_nb_noeuds+1],*ind_pile;
tarbre *Pile[Max_nb_noeuds];
{
/* Parcours l'arbre en profondeur en gerant la pile pr le marquage */
int n,i,v=0;
tarbre *s,*som;
/* Si contient des taxa a atteindre marque les noeuds ds la pile
   depuis le dernier marque de la pile */

som=Pile[(*ind_pile)-1];

for (i=0;som->taxa[i];i++)
  if (Atteindre[som->taxa[i]])
    {
     (*Reste)--;
     n=(*ind_pile)-1; /* marque depuis le sommet jusqu'au premier som
			 marque recontre ds la pile en la descendant */
     while (!marques[Pile[n]->num]) 
       { 
         marques[Pile[n]->num]=1;
         n--;}
    }

/* Parcours ses fils tant que certains taxa restent a joindre */
while ((*Reste>0) && (som->voisins[v]))
  {
   if (((*ind_pile)>1) && (Pile[(*ind_pile)-2]==som->voisins[v]) )
       {;/*printf(" son premier voisin est son pere ds la pile\n");*/}
   else
        {
	 s=som->voisins[v];
	 Pile[*ind_pile]=s;(*ind_pile)++; /* empile s */
	 Profondeur(Reste,Atteindre,marques,ind_pile,Pile); 
	 (*ind_pile)--;Pile[*ind_pile]=NULL; /* depile s */
         }
   v++;
  }
}
/*--------------------------------------------*/
void Parcours(b,cote,marques)
int b,cote,marques[Max_nb_noeuds+1];
/* Met dans "marques" le nom de tous les sommets de l'Arbre
   correspondant au sous-arbre d'un des 2 cotes de la bipartition b */
/* Pour ca effectue un parcours en profondeur depuis un sommet de
   ce cote */
/* Comme la topologie courante est un arbre on a la propriete que les
   voisins d'un sommet issu du noeud i, ne sont pas deja atteints,
   sauf le voisin qui correspond au noeud i justement. */
/* Pile contient ds une dim le num des noeuds et a cote si ils sont
   marques(1) ou pas(0) */
{
int Atteindre[Max_nb_taxa+1];
tarbre *Pile[Max_nb_noeuds];
int ind_pile,Reste_a_Atteindre=0;
int i,taxon_dep;
/*---- Init ---- */
    /* cree le vecteur de taxa a atteindre pour cet partie -> O(1)
       pour savoir plus tard si un noeud donne doit etre marque ou pas */
    for (i=1;i<=Max_nb_taxa;i++) Atteindre[i]=0;
    for (i=0;bipartition[b][cote][i]!=0;i++) 
      {Atteindre[bipartition[b][cote][i]]=1;Reste_a_Atteindre++;}
   /* Init la pile a vide */
   for (i=0;i<Max_nb_noeuds;i++) Pile[i]=NULL;
/*---------------*/

/* On part d'un sommet de Arbre qui est un des som qu'on veut atteindre */
taxon_dep=bipartition[b][cote][0]; /* non vide */
/*printf("depart taxon %d donc du noeud  %d\n",taxon_dep,sommet[taxon_dep]->num);*/
ind_pile=0;
Pile[ind_pile]=sommet[taxon_dep];
marques[sommet[taxon_dep]->num]=1;
ind_pile++;

/* parcours proprement dit */
Profondeur(&Reste_a_Atteindre,Atteindre,marques,&ind_pile,Pile);
}
/*--------------------------------------------*/
int Intersection(gche,dte,noeud)
int gche[Max_nb_noeuds+1],dte[Max_nb_noeuds+1],*noeud;
{
int nb=0,i;
for (i=1;i<=Max_nb_noeuds;i++)
 if ((gche[i]) && (dte[i])) {nb++;*noeud=i;}
return(nb);}
/*--------------------------------------------*/
int Appartient(b,part,taxon)
int b,part,taxon;
{
int trouve=0,t=0;
while ((!trouve) && (bipartition[b][part][t])) 
 {if (bipartition[b][part][t]==taxon) trouve=1;
  t++;}
return(trouve);
}
/*--------------------------------------------*/
void Scinde(nd,b,arbG,arbD)
int nd,b,arbG[Max_nb_noeuds+1],arbD[Max_nb_noeuds+1];
{
/* separe le noeud en 2 et repartit les voisins de l'ancien noeud de
   facon a ce que les sous-arbres associes a chq partie de la
   bipartition en question soient connexes */
tarbre *neosom,*noeud,*vois;
int ind,ivieux,nbvois,i,j,v,vvieux;
tarbre *voisin_Dt[Max_nb_noeuds];
int taxa_Dt[Max_nb_taxa];
noeud=noeuds[nd];
if ((neosom=(tarbre *)malloc(sizeof(tarbre)))==0)
     Erreur("MEMORY ALLOCATION ERROR (no more space available)",2);
neosom->num=Num_prochain_noeud;Num_prochain_noeud++;noeuds[neosom->num]=neosom;
/* lui associe les TAXA de l'ancien appartenant a la partie gche de la
   bipartition : O(n) */
ind=0;ivieux=0;
for (i=0;noeud->taxa[i];i++)
  {
   if (Appartient(b,1,noeud->taxa[i]))
     {neosom->taxa[ind]=noeud->taxa[i];ind++;
      sommet[noeud->taxa[i]]=neosom;}
   else {taxa_Dt[ivieux]=noeud->taxa[i];ivieux++;}
  }

for (j=ind;j<Max_nb_taxa;j++) neosom->taxa[j]=0;           
/* a l'ancien laisse les taxa de la partie dte */
for (i=0;i<ivieux;i++) 
  {noeud->taxa[i] = taxa_Dt[i];}
for (j=i;j<Max_nb_taxa;j++) noeud->taxa[j]=0;

/* lui associe les voisins de l'ancien sommet qui appartiennent au
   sous-arbre associe a la partie gche de la bipartition. Les voisins
   de l'autre sousarbre les met ds voisin_Dt : O(n) */
v=0;
vvieux=0;nbvois=0;
while (noeud->voisins[vvieux])
  {
   if (arbG[noeud->voisins[vvieux]->num])
      {neosom->voisins[v]=noeud->voisins[vvieux];v++;}
   else {voisin_Dt[nbvois]=noeud->voisins[vvieux];nbvois++;} 
   vvieux++;
   }
neosom->voisins[v]=NULL;
/* chge le pointeur des noeuds VOISINS vers le neo noeud : O(n) */
for (v=0;neosom->voisins[v];v++)
  {
   vois=neosom->voisins[v];
   for (i=0;(vois->voisins[i]) && (vois->voisins[i]!=noeud);i++)   ;
   if (!vois->voisins[i]) 
     Erreur("je ne trouve pas l'ancien noeud ds la liste des\nvoisins de ses voisins",4);
   vois->voisins[i]=neosom;
  }

/* garde que les voisins de la sousarb dte dans l'ancien noeud */
for (i=0;i<nbvois;i++) noeud->voisins[i]=voisin_Dt[i];
for (i=nbvois;i<Max_nb_noeuds;i++) noeud->voisins[i]=NULL;

/* ajoute une arete entre le neo et l'ancien noeud */
noeud->voisins[nbvois]=neosom;
neosom->voisins[v]=noeud;v++;neosom->voisins[v]=NULL;

/*printf("les noeuds voisins du neo som sont : ");
for (i=0;neosom->voisins[i];i++) printf("%.2d ",neosom->voisins[i]->num);
printf("\n"); printf("les noeuds voisins de l'ancien som sont : ");
for (i=0;noeud->voisins[i];i++) printf("%.2d ",noeud->voisins[i]->num);
printf("\n");printf("Taxa(noeud) = ");
for (i=1;sommet[i];i++) printf("%.2d(%d) ",i,sommet[i]->num); printf("\n");*/
}

/*_________________________________________________*/
/*____________ fin de l'algo tree-popping _________*/
/*_________________________________________________*/

/*--------------------------------------------*/
void Integre_bipartition(b)
int b;
/* Traite la bieme bipartition */
{
int noeud=-1;
int noeuds_gche[Max_nb_noeuds+1],noeuds_dte[Max_nb_noeuds+1]; 
int i;
/* Init */
for (i=0;i<=Max_nb_noeuds;i++) {noeuds_gche[i]=noeuds_dte[i]=0;}
/* Cherche noeuds (internes et externes) du cote gche de la bip */
Parcours(b,1,noeuds_gche);
/*printf("Sous-arbre gche = noeuds ");
for (i=1;i<=Max_nb_noeuds;i++) 
  if (noeuds_gche[i]) printf("%.2d ",i);  printf("\n");*/

/* Cherche noeuds du cote dt */
Parcours(b,2,noeuds_dte);
/*printf("Sous-arbre dt = noeuds ");
for (i=1;i<=Max_nb_noeuds;i++) 
if (noeuds_dte[i]) printf("%.2d ",i); printf("\n");*/

/* Si intersection == 2 noeuds -> incompatibilite 
                   == 0 noeuds -> rien faire 
		   == 1 noeud -> le separer en 2 */
switch (Intersection(noeuds_gche,noeuds_dte,&noeud))
	{
	case 0:return;break;
	case 2:{Erreur("INCOMPATIBLE BIPARTITIONS !!!!\n",5);}break;
	case 1:{/* recois le "noeud" comme seule intersection ;*/
		Scinde(noeud,b,noeuds_gche,noeuds_dte);}
        }
}
/*--------------------------------------------*/
void Affiche_voisins()
{
int i,j;
for (i=1;noeuds[i];i++)
  {printf("n%d (%.2d) -> ",noeuds[i]->num,noeuds[i]->taxa[0]);
   for (j=0;noeuds[i]->voisins[j];j++)
       printf("n%d ",noeuds[i]->voisins[j]->num);
  printf("\n");
  }
}
/*--------------------------------------------------*/
void Sauve_Arbre(f,noeud,pere)
FILE *f;tarbre *noeud,*pere;
{
int i,prem;
/* Produit un arbre ENRACINE depuis la structure de donnees */
/* affiche ses taxa */
for (i=0;noeud->taxa[i];i++) fprintf(f,"%s",name[noeud->taxa[i]]);
/* recurse */
if (noeud->voisins[1]) /* a un autre voisin que son pere */
    {
     fprintf(f,"(");prem=1;
     for (i=0;noeud->voisins[i];i++)
       {
	  if (noeud->voisins[i]!=pere)  
	    {if (!prem) fprintf(f,",");
	     else prem=0;
	     Sauve_Arbre(f,noeud->voisins[i],noeud);
	    }
	}
     fprintf(f,")");
    }
}
/*--------------------------------------------------*/
int Aff_Agi()
{
int i,j;listequads *q;

for (i=0;i<Nb_noeuds_agreg;i++)
  for (j=0;j<i;j++)
    {printf("ag(n%d,n%d) : ",refinv[i],refinv[j]);
     if (Agi[i][j].Lqd)
       {q=Agi[i][j].Lqd;
	do
	  {printf("%d%d||%d%d ",q->gche[0],q->gche[1],q->dte[0],q->dte[1]);
	   q=q->suiv;}
	while (q!=Agi[i][j].Lqd);} 
     printf("\n");}

printf("\n(sx(x),sy(y),sw(w)) des Agi:\n");
printf("  ");
for (j=0;j<Nb_noeuds_agreg;j++) printf("    %d    ",j);printf("\n");
for (i=0;i<Nb_noeuds_agreg;i++)
  {printf("%d ",i);
   for (j=0;j<i;j++)
     printf(" (%lf(%d),%lf(%d),%lf(%d)) ",Agi[i][j].sx,Agi[i][j].x,Agi[i][j].sy,Agi[i][j].y,Agi[i][j].sw,Agi[i][j].w);
 printf("\n");
 }
}
/*-------------------------------*/
void Quadruplets_multi(num)
int num;
{
listequads *qd[2][3],*elem,*prec,*maillon;
int nbconc1=0,nbconc2=0,b1,b2,b3,b4,app,tmp,d,i,lc1,lc2;
int q[3][4]; /* bon ordonnancement des tax pour chacun des 3 r4-arbres */
TaxList *l1,*l2,*l3,*l4,*e1,*e2,*e3,*e4;
tarbre *multi;
double score;
/* maj les structures pour resoudre la multifourche num */
Correspondance_nds_ligcol(num); 

multi=noeuds[Multifourches[num]];
/* Pour chq combinaison de 4 branches autour de la multi Mi */
for (b1=0,l1=multi->tax_sousarb[b1];multi->voisins[b1+3];b1++,l1=multi->tax_sousarb[b1])
 for (b2=b1+1,l2=multi->tax_sousarb[b2];multi->voisins[b2+2];b2++,l2=multi->tax_sousarb[b2])
  for(b3=b2+1,l3=multi->tax_sousarb[b3];multi->voisins[b3+1];b3++,l3=multi->tax_sousarb[b3])
     for(b4=b3+1,l4=multi->tax_sousarb[b4];multi->voisins[b4];b4++,l4=multi->tax_sousarb[b4])
    /* pr chq combin de tax appartenant chacun a une des 4 branches choisies */
    /* ie, pour chq quad susceptible d'etre resolu par une agglomeration */
       for (e1=l1->suiv;e1;e1=e1->suiv)
	 for (e2=l2->suiv;e2;e2=e2->suiv)
	   for (e3=l3->suiv;e3;e3=e3->suiv)
	     for (e4=l4->suiv;e4;e4=e4->suiv)
	       {/*printf("%d%d||%d%d ",e1->num,e2->num,e3->num,e4->num);*/
		/* ordonne correctement les tax des 3 r4-arbres */
		 /* 1er r4-arbre */
	        if (e1->num<e2->num) {q[0][0]=e1->num;q[0][1]=e2->num;}
		else                 {q[0][0]=e2->num;q[0][1]=e1->num;}
		if (e3->num<e4->num) {q[0][2]=e3->num;q[0][3]=e4->num;}
		else                 {q[0][2]=e4->num;q[0][3]=e3->num;}
		if (q[0][0]>q[0][2]) {tmp=q[0][0];q[0][0]=q[0][2];q[0][2]=tmp;
				      tmp=q[0][1];q[0][1]=q[0][3];q[0][3]=tmp;}
		/* 2eme r4-arbre */
	        if (e1->num<e3->num) {q[1][0]=e1->num;q[1][1]=e3->num;}
		else                 {q[1][0]=e3->num;q[1][1]=e1->num;}
		if (e2->num<e4->num) {q[1][2]=e2->num;q[1][3]=e4->num;}
		else                 {q[1][2]=e4->num;q[1][3]=e2->num;}
		if (q[1][0]>q[1][2]) {tmp=q[1][0];q[1][0]=q[1][2];q[1][2]=tmp;
				      tmp=q[1][1];q[1][1]=q[1][3];q[1][3]=tmp;}
		/* 3eme r4-arbre */
	        if (e1->num<e4->num) {q[2][0]=e1->num;q[2][1]=e4->num;}
		else                 {q[2][0]=e4->num;q[2][1]=e1->num;}
		if (e2->num<e3->num) {q[2][2]=e2->num;q[2][3]=e3->num;}
		else                 {q[2][2]=e3->num;q[2][3]=e2->num;}
		if (q[2][0]>q[2][2]) {tmp=q[2][0];q[2][0]=q[2][2];q[2][2]=tmp;
				      tmp=q[2][1];q[2][1]=q[2][3];q[2][3]=tmp;}
		/* Est-ce qu'une des 3 r4-arbres lies a ce quad est ds Q ? */
		/* Pour recuperer le score du quad on utilise le fait que 
		   le test s'arrete des qu'un des termes du ou est vrai */ 
		app=0;
		if   (((score=MatQ[q[0][0]][q[0][1]][q[0][2]][q[0][3]]) 
		    || (score=MatQ[q[1][0]][q[1][1]][q[1][2]][q[1][3]]))
		    || (score=MatQ[q[2][0]][q[2][1]][q[2][2]][q[2][3]])) 
		  app=1;
		/* else printf("quadruplet pas resolu ds donnees\n");*/
		/* Cree les 6 r4-arbres */
		for (d=0;d<2;d++)
		  for (i=0;i<3;i++)
		  {
		    if ((qd[d][i]=(listequads *)malloc(sizeof(listequads)))==0)
		       {printf("MEMORY ALLOCATION ERROR (no more space available)\n");exit(-3);}
		    if (!d) /* pr inverser les 2 couples 1 fois sur 2 */
		            /* par ex (ab)|(cd) puis (cd)|(ab)  */
		      {qd[d][i]->gche[0]=q[i][0];qd[d][i]->gche[1]=q[i][1];
		       qd[d][i]->dte[0]=q[i][2];qd[d][i]->dte[1]=q[i][3];}
		    else 
		      {qd[d][i]->dte[0]=q[i][0];qd[d][i]->dte[1]=q[i][1];
		       qd[d][i]->gche[0]=q[i][2];qd[d][i]->gche[1]=q[i][3];}
		    /* rattache chacun a son Agi */
		    Quelle_Agi(qd[d][i],&lc1,&lc2);
		    if (Agi[lc1][lc2].Lqd)
		      {qd[d][i]->prec=Agi[lc1][lc2].Lqd->prec;
		       qd[d][i]->prec->suiv=qd[d][i];
		       Agi[lc1][lc2].Lqd->prec=qd[d][i];
		       qd[d][i]->suiv=Agi[lc1][lc2].Lqd;}
		    else {
		      qd[d][i]->prec=qd[d][i]->suiv=qd[d][i];
		      Agi[lc1][lc2].Lqd=qd[d][i];}
		    /* maj le score de cette Agi */
		    if (!app) {qd[d][i]->typ='w';
			       (Agi[lc1][lc2].w)++;
			       /* donne arbitrairement une penalite de 1 
				  pour la resolution de ce quadruplet */
			       (Agi[lc1][lc2].sw)++;}
		    else 
		      {if (MatQ[q[i][0]][q[i][1]][q[i][2]][q[i][3]])
			 {qd[d][i]->typ='x';
			  (Agi[lc1][lc2].x)++;(Agi[lc1][lc2].sx)+=score;}
		      else {qd[d][i]->typ='y';
			    (Agi[lc1][lc2].y)++;
			    /* donne une penalite pour resoudre ce quadruplet
			       equivalente au score de son bon 4-arbre */
			    (Agi[lc1][lc2].sy)+=score;
			  }}
		  }
		/* les lie entre eux */
		qd[0][0]->qdsuiv=qd[0][1];qd[0][1]->qdsuiv=qd[0][2];
		qd[0][2]->qdsuiv=qd[1][0];qd[1][0]->qdsuiv=qd[1][1];
		qd[1][1]->qdsuiv=qd[1][2];qd[1][2]->qdsuiv=qd[0][0];
	      }
}
/*-------------------------------*/
int Compte_sat_reletoile()
{
FILE *fq;
int t,nb1=0,tmp,combien;
char s[50],unquad[50];
if (! (fq=fopen("quartfile.res","r")) ) 
   return(0); /* ce fichier n'existe pas */

/* Lit le nb de taxa */
fscanf(fq,"%d",&combien);
if (combien!=Nb_taxa) Erreur("INCOHERENT NUMBER OF TAXA BETWEEN bipfile AND quartfile.res FILES",6);
/* Lit la correspondance numero-espece */
for (t=1;t<=Nb_taxa;t++)
  fscanf(fq,"%d %s\n",&tmp,&s);
fscanf(fq,"\n");
/* Lit les quadruplets */
for (nb1=0;fgets(unquad,50,fq);nb1++) ;
printf("%d quartets already resolved in the QSTAR tree\n",nb1);
fclose(fq);
return(nb1);
}
/* -------------------------------------------- */
void Aff_liste(deb)
TaxList *deb;
{TaxList *p;
for (p=deb->suiv;p;p=p->suiv)
{printf(" %d ",p->num);}
printf("\n");
}

/* -------------------------------------------------- */
void Taxa_Sous_arbo(pere,nd,listaxa)
tarbre *pere,*nd; TaxList *listaxa;
/* Determine pour chq noeud (sauf les externes) , quelles sont les
taxa contenus dans la sous-arbo induite par chacun de ses voisins.
Renvoie l'union (ie, les taxa de la sousarbo nd) dans "listaxa"*/
/* SI : listaxa contient un bidon pour pouvoir recuperer le res */
{
int t,fils,indpere,contient_pere[Max_nb_taxa];
TaxList *lfils,*neo;
/* Est-ce qu'on est sur un noeud externe (un seul voisin: son pere)*/
if (!nd->voisins[1]) /* alors renvoie le taxon */
  {/*printf("je suis sur un noeud externe : n%d\n",nd->num);*/
   nd->tax_sousarb[0]=NULL;
   if ((neo=(TaxList *)malloc(sizeof(TaxList)))==0)
     {printf("MEMORY ALLOCATION ERROR (no more space available)\n");
      exit(-3);}
   neo->num=nd->taxa[0]; /* on sait que now un seul taxa */
   neo->suiv=listaxa->suiv;listaxa->suiv=neo;}
else /* recupere les taxa de chq sousarbo associee a un fils */
  {/*printf("je suis sur un noeud interne : n%d\n",nd->num);*/
    for (t=1;t<=Nb_taxa;t++) contient_pere[t]=1;
    /* pr chq sousarbo fille */
    for (fils=0;nd->voisins[fils];fils++)
      if (nd->voisins[fils]!=pere)
	{/*printf("je vais voir son fils : n%d\n",nd->voisins[fils]->num);*/
	 if ((nd->tax_sousarb[fils]=(TaxList *)malloc(sizeof(TaxList)))==0)
	     Erreur("MEMORY ALLOCATION ERROR (no more space available)",2);
         nd->tax_sousarb[fils]->suiv=NULL;
	 Taxa_Sous_arbo(nd,nd->voisins[fils],nd->tax_sousarb[fils]);
         /*printf("Je reviens avec :");Aff_liste(nd->tax_sousarb[fils]);*/
	 /* Union-recopie avec listaxa + enleve de contient_tax pere */
	 for (lfils=nd->tax_sousarb[fils]->suiv;lfils;lfils=lfils->suiv)
	   {
	    if ((neo=(TaxList *)malloc(sizeof(TaxList)))==0)
	     Erreur("MEMORY ALLOCATION ERROR (no more space available)",2);
	    neo->num=lfils->num;neo->suiv=listaxa->suiv;listaxa->suiv=neo;
	    contient_pere[lfils->num]=0;
	   }
        }
      else indpere=fils;
    /* deduit la sous-arbo du pere par {1..n}-union(sousarbo(fils)) */
    if ((nd->tax_sousarb[indpere]=(TaxList *)malloc(sizeof(TaxList)))==0)
      {printf("MEMORY ALLOCATION ERROR (no more space available)\n");
       exit(-3);}
    nd->tax_sousarb[indpere]->suiv=NULL;
    for (t=1;t<=Nb_taxa;t++) 
      if (contient_pere[t])
	{ if ((neo=(TaxList *)malloc(sizeof(TaxList)))==0)
	     {printf("MEMORY ALLOCATION ERROR (no more space available)\n");
	      exit(-3);}
	  neo->num=t;
	  neo->suiv=nd->tax_sousarb[indpere]->suiv;
	  nd->tax_sousarb[indpere]->suiv=neo;
	}
   /*printf("J'en deduit que le pere contient les espces ");
   Aff_liste(nd->tax_sousarb[indpere]);*/
  }
}
/*_______________________________________________________*/
/*--------------------------------------------------*/
void Especes_Sous_arbo(pere,racine,ct)
tarbre *racine,*pere; int *ct;
/* Cherche les esps de la sous-arbo partant de "racine" definie par la
   direction opposee a "pere", et remplit la struct "contient_tax"
   depuis l'indice "ct"*/
{
int fils,tax;
/* ajoute les eventuels taxa de ce noeud */
for (tax=0;racine->taxa[tax];tax++)
  {contient_tax[*ct]=racine->taxa[tax];(*ct)++;}
/* parcours la sous-arbo de racine */
for (fils=0;racine->voisins[fils];fils++)
  if (racine->voisins[fils]!=pere)
    Especes_Sous_arbo(racine,racine->voisins[fils],ct);
}
/*--------------------------------------------------*/
int main(argc, argv)
int argc;char *argv[];
{extern char *optarg;
char c,nomfic[20];
int nbquads=0;
int i,j,k,cpt,quads_deja_sat,star_tree=0;
FILE *farb;
Nb_taxa=4;
/* Process parameters */
while ((c = getopt(argc, argv, "a:t:f:")) != -1)
     switch (c) {
     case 'a':a=atof(optarg);break;
     case 't':star_tree=atoi(optarg);break;
     case 'f':fully_resolved=atoi(optarg);break;
     case '?':Erreur("UNKNOWN PARAMETER !",1);
                 }
/*---------------------------------------------------------------*/
/* 1 : Lit les quadruplets donnes en entree (non-resolus par Q*) */
/*---------------------------------------------------------------*/
Lit_quad_restant(star_tree,&nbquads); /* les place ds le tabl a 4 dim MatQ */
printf("%d quartets read in input file.\n",nbquads);
/*-----------------------------------------------------------------------------*/
/* 2 : Lit les bipartitions eventuellement donnes en entree (proposees par Q*) */
/*-----------------------------------------------------------------------------*/
Lit_bipartitions(star_tree);
/*--------------------------------------------------------------------*/
/* 2-suite : creer l'arbre depuis les bipartitions ||* : tree-popping */
/*--------------------------------------------------------------------*/
/*--- Init arbre avec ts les taxa ds un noeud --*/
if ((Arbre=(tarbre *)malloc(sizeof(tarbre)))==0)
     Erreur("MEMORY ALLOCATION ERROR (no more space available)",2);
for (i=0;i<Nb_taxa;i++) {Arbre->taxa[i]=i+1;sommet[i+1]=Arbre;}
for (i=Nb_taxa;i<Max_nb_taxa;i++) {Arbre->taxa[i]=0;sommet[i+1]=NULL;}
for (i=1;i<=Max_nb_noeuds;i++) noeuds[i]=NULL;
Arbre->num=1;Num_prochain_noeud=2;noeuds[1]=Arbre;
Arbre->voisins[0]=NULL;

/* Incorpore les bipartitions une a une */
for (i=1;i<=Nb_Bips;i++) Integre_bipartition(i);

/* Etablit list des taxa des sous-arbre par parcours recursif
   depuis le taxon 1 */
if ((sommet[1]->tax_sousarb[0]=(TaxList *)malloc(sizeof(TaxList)))==0)
     Erreur("MEMORY ALLOCATION ERROR (no more space available)",2);
sommet[1]->tax_sousarb[0]->suiv=NULL;
Taxa_Sous_arbo(sommet[1],sommet[1]->voisins[0],sommet[1]->tax_sousarb[0]);   
/*----------------------------------------------------------*/
/* 3 : etablit la liste des multifourches : O(n) 
       se fait simplement en parcourant la struct noeuds en
       cherchant les noeuds connectes a plus de 3 autres    */
/*----------------------------------------------------------*/
/* RAZ */
for (i=0;i<Max_nb_noeuds;i++) Multifourches[i]=0;
i=1;
while (noeuds[i])
  {
    if (  ((noeuds[i]->voisins[2])&&(noeuds[i]->voisins[3]))
       && ((noeuds[i]->voisins[0])&&(noeuds[i]->voisins[1])) )
      {Multifourches[nb_multi]=i;nb_multi++;}
    i++;
  }

/* printf("Arbre ||* = (01,");  
   Affiche_Arbre(sommet[1]->voisins[0],sommet[1]);printf(");\n"); */

fbip=fopen("bipfile","a"); /* ajoute bips creees dans bipfile */
/*------------------------------------------------------------*/
/* 4 : Initialize various counters of resolved quartets       */
/*------------------------------------------------------------*/
Critere_global.x=0;
Critere_global.y=0;
Critere_global.w=0;

if (!star_tree)
  {
    quads_deja_sat=Compte_sat_reletoile();
    Critere_global.x+=quads_deja_sat;
    nbquads+=quads_deja_sat;
  }
/*--------------------------------------------------*/
/* 5 : Resoud chq multifourche tour a tour          */
/*--------------------------------------------------*/
srandom(getpid()); 
printf("\nResolving multifurcations :\n");
for (i=0;i<nb_multi;i++)
  {
   /* Etablir la liste des esps par sous-arbo % multi 
      pas faisable avant car 2 nds multi voisins peuvent ne plus 
      etre voisins apres resolution d'un des 2 */ 
   /*printf("n%d ->  ",Multifourches[i]);*/
   for(j=0;noeuds[Multifourches[i]]->voisins[j];j++)
     {
       for(k=0;k<Nb_taxa;k++) contient_tax[k]=0;
       cpt=0;
       Especes_Sous_arbo(noeuds[Multifourches[i]],noeuds[Multifourches[i]]->voisins[j],&cpt);
     }
    /* printf("INIT POSSIBLE AGREGATIONS\n");*/
    Quadruplets_multi(i);
    /* FAIT L'ALGO "GLOUTON" POUR CETTE MULTIFOURCHE */
    /* printf("RESOLVING A MULTIFURCATION\n");*/
    Resoud_multi(i);
 }
fclose(fbip);

/* On obtient ainsi un arbre pas forcement resolu et non-enracine, */
/* qu'on se contente d'afficher pour le moment */
/*printf("Arbre apres resolution des multifourches par glouton = \n");
printf( "(01," );
Affiche_Arbre(sommet[1]->voisins[0],sommet[1]);printf(");\n");*/

printf("\n%d (%.2f%%) input quartets are satisfied by the inferred tree\n",Critere_global.x,100.0*(float)Critere_global.x/(float)nbquads);
printf("%d (%.2f%%) input quartets are contradicted by the inferred tree\n",Critere_global.y,100.0*(float)Critere_global.y/(float)nbquads); 
printf("%d (%.2f%%) quartets unresolved in the input are however resolved by the inferred tree\n",Critere_global.w,100.0*(float)Critere_global.w/(float)nbquads);

farb=fopen("treefile","w");
fprintf(farb,"(%s,",name[1]);
Sauve_Arbre(farb,sommet[1]->voisins[0],sommet[1]); /* enracine au taxon 1 */
fprintf(farb,");\n");
fclose(farb);

}






