/***************************************************************/
/*  QSTAR : version of 5/8/99 - this program is part of the 
            PHYLOQUART package v1.4  .                        
    (vberry@lirmm.fr)                                         

    Please distribute with the full package                    */
/***************************************************************/
/* Program extracting the Q* set in O(n^4) time from a set Q of 
   input quartet relationships among a set of taxa.
   I.e., infers the maximum set of bipartitions on taxa s.t.
   all quartet relationships induced by these bips are present
   in the input quartet set.
   For more details I refere the reader to 
              [Berry & Gascuel , RECOMB 1997]
              [Berry & Gascuel , Theor. Comp. Science 2000]    
   available on simple request addressed to vberry@lirmm.fr    

   See also [Bunemen71,Benzecri73,BandeltDress86]              */
/***************************************************************/
/* INPUT  = Set Q of quartet relationship in file "quartfile"  */
/* OUTPUT = Set of inferred bipartitions in file "bipfile"     */
/*          Quartets of Q* in file "quartfile.res"             */
/*          Remaining quartets in Q in file "quartfile.left"   */
/* Compx = O(n^4)                                              */
/* Note: in current version, if weighted quartets are input, 
   ignore weights     */
/***************************************************************/

/* TODO list :

   Passer les variables en static ds la version distribuee 

OPTIMISATIONS !!!!!!!!!!!!!!!  :

Lors du 2eme parcours (descendant prefixe), d\`es qu'on rencontre une
arete e_i pour laquelle l'etiquette calculee lors du 1er parcours est
"oui" (i.e., i peut etre insere au dessus), alors se contenter de
mettre les etiquettes a "non" dans toute la sous-arbo inferieure de
e_i, SANS MEME REGARDER LES QUADRUPLETS DE Q OU METTRE A JOUR LES
STRUCTURES Z (on en aura pas besoin pour un parcours futur). Par contre 
on ne peut pas faire ca lors du 1er parcours car il faut que les struct
Z soient a jour puisque le 2eme parcours en a besoin).

Justification : si i peut etre insere au dessus de e_i, il ne pourra
pas etre insere au-dessous. 

Autre OPTIMISATION : ne pas considerer les especes dans n'importe quel
ordre, par exemple, on peut lors d'un examen de Q compter pour chaque
espece le nb de 4-arbres dans lesquels elle est impliquee (et aussi 
peut-etre avec quelles autres esp elle est le plus souvent voisine).

Ca devrait donner des infos pour savoir un peu dans quel ordre il
vaut mieux inserer les especes. Le mieux est peut-etre d'inserer avant tout
les especes les plus frequentes, de facon a avoir des Qi de taille reduite
quand l'arbre T* commence a avoir plus d'aretes ? et l'inverse ?
Faire des essais, pour savoir ce qui est le plus rapide.

Autre OPTIMISATION : Lors du 2eme parcours, avant de remonter dans
l'arbre (i.e. fin de traitement d'une arete), on peut en
profiter pour supprimer l'arete si elle est "non"x2 !? 

Autre OPTIMISATION : binariser artificiellement l'arbre, car rien
n'est plus couteux que la topologie etoile : on a du O(n^4) pour tout
ce qui est gestion des structures Z, ... alors que si l'arbre est
binaire, on a du O(n^3) pour les memes choses (on n'a plus que du
O(n^4) lie au nombre de 4-arbres examines dans Q).
Toutefois, est-ce que ca amene un GAIN EN TEMPS D'EXECUTION ?
Bien reflechir d'abord en fonction des struct actuelles 

Peut-etre que cette idee ne vaut le coup qu'au dela d'un certain nbre 
d'especes ? faire des essais et parametrer le programme par un simple
test en debut.
*/

#include <stdio.h>
#include <stdlib.h>

#define Max_nb_taxa 400 /* 477 largest n s.t. C(n,4) < 2^31 */

#define Inf 0
#define Sup 1
#define Prefixe 0
#define Postfixe 1
#define Non 0
#define Oui 1

typedef struct tliste {int taxon; struct tliste *suiv;} tliste;
/* L'arbre T_i est gere par pointeurs, chaque noeud et chaque arete
   etant une entite distincte. L'arbre est enracine a l'espece 1 */

typedef struct tarete    {
                          struct tnoeud *fils,*pere;
			  tliste *X[2]; /* Inf et Sup */
			  int Z[2][Max_nb_taxa+1]; /* on a parfois besoin de
						 Zj pour Inf et Sup a la fois */
			  int horsde[2]; /* 0=non 1=oui */
                         } tarete;

typedef struct listarete { tarete *a; struct listarete *suiv;} listarete ;

typedef struct tnoeud    {
                         int taxon; /* num du taxon ou num nd interne */
                         struct tarete *haut;
                         listarete *bas,*fin; /* pteur sur dernier elem pr union en O(1) */  
                         } tnoeud;


/* --------------Var globales-------------- */
int Nb_taxa;
unsigned long NbQuartRead,NbQuartRes=0;

tnoeud *Taxon1; /* la racine de l'arbre */
int Num_intern,Etape;
char name[Max_nb_taxa+1][30];
short *up_part, *down_part; /* for saving bipartitions */
FILE *fr,*fl,*fb;

/*---------------------------------------------------------------------------------------*/
/* New struct for storing quartets */
#define Tquart unsigned char /* just 1 byte */
/* 4 quartets are coded in a byte */
/* Each quartet can have 4 values : 
   00 = unresolved
   01 = ij||kl
   10 = ik||jl
   11 = il||jk */
/* Declaration de la struct contenant les quadruplets */
Tquart **Quartets; 
/* 

   ATTENTION : 

  now les nums d'especes commencent a 0 et non plus a 1 !!!!!!!!!!!!
   pbm correspondance des names[] ? et sauvegarde des quads,bips ?????? 



*/
/*---------------------------------------------------------------------------------------*/
/*int Max_nb_bips = 2*Nb_taxa - 3;*/
int Nb_bips=0,bips[2*(Max_nb_taxa+1)-3][2][Max_nb_taxa+1]; /* taxa pas ranges
							 dans le bon ordre */
/*---------------------------------------------------------------*/
void Erreur(mess,valerr)
char *mess;int valerr;
{
printf("\n\n\tERROR: %s\n\n",mess);
printf("Please send a bug report to vberry@lirmm.fr for explanations/corrections\n");
exit(valerr);
}
/*---------------------------------------------------------------*/
void Affiche_Arbre(n)
tnoeud *n;
{
listarete *Sj;
for (Sj=n->bas;Sj;Sj=Sj->suiv)
  {printf("%d --(%d|%d)-- %d\n",n->taxon,Sj->a->horsde[Sup],Sj->a->horsde[Inf],Sj->a->fils->taxon);
   Affiche_Arbre(Sj->a->fils);}
}

/*-----------------------------------------*/
void Affiche_Xj(Sj,sens)
tarete *Sj;int sens;
{tliste *x;
for (x=Sj->X[sens];x;x=x->suiv) printf(" %d",x->taxon); 
printf(".\n");
} 
/*---------------------------------*/
void Order4tree(x,y,z,t)
int *x,*y,*z,*t;
{  /* Order species s.t. x<y , z<t and x<z */
int tmp;
if (*x>*y) {tmp=*x;*x=*y;*y=tmp;} 
if (*z>*t) {tmp=*z;*z=*t;*t=tmp;}
if (*z<*x) {tmp=*x;*x=*z;*z=tmp; tmp=*y;*y=*t;*t=tmp;}
}
/*-----------------------------------------*/
void InitQuartets()
{
int nij,nkl,i,j,ij,nbbytes;
nij = (Nb_taxa - 3)*(Nb_taxa - 2)/2;
if (!(Quartets=(Tquart **) calloc (nij,sizeof(Tquart *))))  /* calloc sets memory to 0 : */
    Erreur("MEMORY ALLOCATION ERROR (no more space available?)",1);
/*printf("VERIFIES THAT calloc SETS MEM TO 0 \n");*/
/*for (i=0;i<nij;i++) 
 printf("%d",Quartets[i]);
 printf("\n");*/

for (j = 1, ij = 0; j < Nb_taxa - 2; j++)
   {
     nkl = (Nb_taxa - j - 1)*(Nb_taxa - j - 2)/2;
     if (nkl % 4) nbbytes = nkl/4+1; else  nbbytes = nkl/4; 
     /*printf("nkl=%d nbbytes=%d\n",nkl,nbbytes);*/
     for (i = 0; i < j; i++, ij++)
       {
	 if (!(Quartets[ij]=(Tquart *) calloc (nbbytes,sizeof(Tquart)) )) /* calloc sets memory to 0 */
	   Erreur("MEMORY ALLOCATION ERROR (no more space available?)",1);
	 /* VERIF QUE ZONE ALLOUEE EST RAZ */
         /*for (tmp=0;tmp<nbbytes;tmp++) printf("%x",Quartets[ij][tmp]); printf("\n");*/
       }
   }
 }
/*-----------------------------------------*/
int SetQuartet(i,j,k,l,resolved)
int i,j,k,l,resolved;
{
/* If resolved = 0 : set quartet relationship to be unresolved */
/* otherwise set the quartet relationship to be : i j || k l */
/* Returns 1 except when setting to unresolved and the quartet being already
   unresolved (for Save_Crossed_Quartet) in which case 0 is returned */
unsigned char value,razbits,mask; 
int ij,kl,kleme,decalage,tmp;
/* Rewrite the quartet such that i < j,k,l and k < l */
 if (i>j) {tmp=i;i=j;j=tmp;} 
 if (k>l) {tmp=k;k=l;l=tmp;} 
 if (i>k) {tmp=i;i=k;k=tmp;tmp=j;j=l;l=tmp;}
 /*printf("SetQ %d,%d || %d,%d",i,j,k,l);*/
/* Encode quartet relationship in "value" by 2 bits : 
   01 = ij||kl  ( 00000001 = 01 hexa ) 
   10 = ik||jl  ( 00000010 = 02 hexa ) 
   11 = il||jk  ( 00000011 = 03 hexa ) 
   00 = unresolved   - set by default by calloc but also here to
                       distinguish quartets resolved by the inferred tree */
if (resolved)
  {if (j<k) value = 0x01;else  if (j<l) value = 0x02; else value = 0x03; }
else value = 0x00;
/*printf("--> %.2x",value);*/
/* Ordonne i,j,k,l par ordre croissant pour trouver la place dans struct quartets */
if (j>k)
  {tmp=j;j=k;k=tmp; /* swap(j,k) */ 
   if (k>l) { tmp=k;k=l;l=tmp;/* swap (k,l) */}}
/* Species stored from 0 to Nb_taxa-1 in quartet struct */
i--;j--;k--;l--;
/* Find place of the corresponding byte in the struct */
ij=j*(j-1)/2 + i;
kleme = (l-j-1)*(l-j-2)/2 + k-j-1;
kl=kleme/4; 
/*printf("(ij=%d,kleme=%d,kl=%d) ",ij,kleme,kl);
  printf("avant modif : %2x ",Quartets[ij][kl]);*/
/* selectionne les bits selon le reste de la div par 4 (4 quads par byte) */
decalage = (3- (kleme % 4))*2;
/*printf("decal = %d",decalage);*/
/* When removing quad from Q (ie resolved=0) test whether quartet is already unresolved */
if (!resolved) 
  {
    /* met les 2 bits concernes a dte et raz le reste */
    mask = (Quartets[ij][kl] >> decalage) & 0x03 ; 
    if (mask==0x00) return(0); /* if unresolved returns 0 */
  }
switch (decalage) {
 case 0: razbits = 0xfc;break; /* razbits is used to set to 0 */
 case 2: razbits = 0xf3;break; /* the 2 interesting bits in the byte addressed */
 case 4: razbits = 0xcf;break; /* before setting their new value. */
 case 6: razbits = 0x3f; }     /* razbits contains 0 for these bits and 1 for others */
Quartets[ij][kl] = ( (Quartets[ij][kl] & razbits ) | (value << decalage ) ); 
/*printf(" apres modif : %2x\n",Quartets[ij][kl]);*/
return(1);
}
/*---------------------------------------------------------------------------*/
/* Returns 1 if quartet is resolved 0 if unresolved.
   If quartet is resolved, returns 'Ftree' to indicate the 4-tree (among the 3) :
      a split occurs between the first and last pairs of taxa 
      (each pair is returned sorted in ascending order) */

int GetQuartet(i,j,k,l,Ftree)
int i,j,k,l,*Ftree;
{
int ij,kl,kleme,decalage,swp;
unsigned char value;

/* Ordonne i,j,k,l par tri fusion */
 Order4tree(&i,&j,&k,&l); /* on devoye l'usage de cette fonction */
 if (k<j)                 /* puis on finit a la main */
   if (l<j) {swp=j;j=k;k=l;l=swp;} 
   else {swp=j;j=k;k=swp;}

/* Species stored from 0 to Nb_taxa-1 in quartet struct */
i--;j--;k--;l--;
/* Find place of the corresponding byte in the struct */
ij=j*(j-1)/2 + i;
kleme = (l-j-1)*(l-j-2)/2 + k-j-1;
kl=kleme/4; 
/*printf("(ij=%d,kleme=%d,kl=%d) ",ij,kleme,kl);*/
value=Quartets[ij][kl];decalage = (3- (kleme % 4))*2;
/*printf("value of the byte %2x decal : %d",value,decalage);*/ 
value = ( (value >> decalage ) & 0x03); /* recupere les 2 bits en question */
/*printf(" 4-tree = %2x\n",value);*/
Ftree[0] = i;
switch (value)
  {
  case 0x00:{return(0);break;}
  case 0x01:{Ftree[1] = j;Ftree[2] = k;Ftree[3] = l;break;}/* Quartet relationship is ij || kl */
  case 0x02:{Ftree[1] = k;Ftree[2] = j;Ftree[3] = l;break;}/* Quartet relationship is ik || jl */
  case 0x03:{Ftree[1] = l;Ftree[2] = j;Ftree[3] = k;break;}/* Quartet relationship is il || jk */
  }
/* Species stored from 0 to Nb_taxa-1 in quartet struct */
for (swp=0;swp<4;swp++) Ftree[swp]++;
/*printf("je renvoie : ");for (swp=0;swp<4;swp++) printf("%d ",Ftree[swp]);   printf("\n");*/
}
/*-----------------------------------------*/
void Lit_quadruplets()
{
FILE *fquad;
char chaine[100];
int t1,t2,t3,t4,tmp,nblus,scoreB=0;
double score;
NbQuartRead=0;
if (!(fquad=fopen("quartfile","r"))) Erreur("READING FILE quartfile",2);

/* Lit le nombre de taxa */
fscanf(fquad,"%d",&Nb_taxa);
printf("%d taxa\n",Nb_taxa);
if (Nb_taxa>Max_nb_taxa) {sprintf(chaine,"Sorry, this program is currently limited to %d taxa",Max_nb_taxa);
                          Erreur(chaine,3);}
/* Init struct for storing quartets */
InitQuartets();
/* Lit la correspondance numero-espece */
for (t1=1;t1<=Nb_taxa;t1++)
  fscanf(fquad,"%d %s\n",&tmp,name[t1]);
fscanf(fquad,"\n");
/*printf("Taxa :\n");
for (t1=1;t1<=Nb_taxa;t1++) printf("%s\n",name[t1]); printf("\n");*/

printf("Reading input quartets ");
while (fgets(chaine,100,fquad))
  {
   /* nouveau quadruplet */
   NbQuartRead++;
   /* extrait les 4 especes et le score si il est present */
   nblus=sscanf(chaine,"%d %d || %d %d (%lf)",&t1,&t2,&t3,&t4,&score);
   /*printf("j'ai lu %d variables : %d %d || %d %d",nblus,t1,t2,t3,t4);*/
   if (nblus==5) scoreB=1; 
   /* Enregistre le quad sans son score */
   SetQuartet(t1,t2,t3,t4,1); /* t1 t2 || t3 t4 */
  }
printf(" : %ld - ok.\n",NbQuartRead);
if (scoreB) printf("WARNING: input quartets are weighted, this program will ignore the weights\n");
}
/*---------------------------------*/
int Appartient_Q(x,y,z,i)
int x,y,z,i;
/* regarde si le 4-arbre xy||zi est dans Q */
{int list[4];
/* Order species s.t. x<y , z<i and x<z */
Order4tree(&x,&y,&z,&i);
if (!GetQuartet(x,y,z,i,list)) return(Non);
/* we know x=list[0];*/
if (y==list[1]) return(Oui); else return(Non);
}
/*--------------------------------*/
tliste *Neo_elem(tax)
int tax;
{tliste *t;
t=(tliste *)malloc(sizeof(tliste));
t->taxon=tax;t->suiv=NULL;return(t);
}
/*---------------------------------*/
void New_fils(nd,typ,num,ou)
tnoeud *nd;char typ;int num;char ou; 
{/* Ajoute un fils de 'typ' interne ('I') ou externe ('E') au noeud
    'nd'. Ce neo noeud est numerote 'num' et est rajoute au debut
    (ou='d') ou a la fin (ou='f') des fils de nd. On cree aussi
    l'arete entre les 2 noeuds. */
int obj;listarete *neo;
if (nd==NULL) Erreur("Tree representation error : invalid father node\n",4);
/*printf("nd =%d, typ = %c, num = %d\n",nd->taxon,typ,num);*/
/* cree l'arete */
neo=(listarete*)malloc(sizeof(listarete));
neo->a=(tarete*)malloc(sizeof(tarete));
neo->a->pere = nd;
neo->a->horsde[Inf]=neo->a->horsde[Sup]=-1;
for (obj=1;obj<Etape;obj++) 
  neo->a->Z[Inf][obj] = neo->a->Z[Sup][obj] = 0;
neo->a->X[Sup]=NULL;
if (typ=='E') neo->a->X[Inf]=Neo_elem(num);
else          neo->a->X[Inf]=NULL;
/* cree le noeud fils */
neo->a->fils=(tnoeud*)malloc(sizeof(tnoeud));
neo->a->fils->haut = neo->a ;
neo->a->fils->taxon=num;
if (typ=='I') Num_intern++;  /* si interne incremente le cptr */
neo->a->fils->bas = neo->a->fils->fin = NULL;

if (ou=='f')
  {/* rattache le tout a 'nd' en fin */
    neo->suiv=NULL;
    if (!nd->bas) nd->bas = neo; /* si 1er fils */
    else nd->fin->suiv=neo; 
    nd->fin = neo;
  }
else 
  {/* rattache le tout a 'nd' en debut */
    if (!nd->bas)  /* si 1er fils */
         {neo->suiv=NULL; nd->bas = nd->fin = neo;}
    else {neo->suiv=nd->bas; nd->bas = neo;}
  }
}
/*---------------------------------*/
void Init_Arbre()
{
tnoeud *n;
int list[4];
/* Init l'arbre T_i depuis la resolution du quadruplet (1,2,3,4) */
/* Creer manuellement le taxon 1 qui sert de racine a l'arbre */
Taxon1=(tnoeud*)malloc(sizeof(tnoeud));
Taxon1->taxon=1;Taxon1->haut=NULL;
New_fils(Taxon1,'I',Num_intern,'f'); /* garder 'I' car init faites */
/* Cas particulier : la racine est au-dessus */
Taxon1->bas->a->X[Sup]=Neo_elem(1);
n = Taxon1->bas->a->fils ; /* on y descend */
if (GetQuartet(1,2,3,4,list))
    switch (list[1]) /* qui est lie avec 1 */
    {
    case 2:{ /* 12||34 */
      New_fils(n,'E',2,'f'); 
      New_fils(n,'I',Num_intern,'f');
      n = n->fin->a->fils ; /* on y descend */
      New_fils(n,'E',3,'f');New_fils(n,'E',4,'f');
      ;break;}
    case 3:{ /* 13||24 */
      New_fils(n,'E',3,'f'); 
      New_fils(n,'I',Num_intern,'f');
      n = n->fin->a->fils ; /* on y descend */
      New_fils(n,'E',2,'f');New_fils(n,'E',4,'f');
      ;break;}
    case 4:{ /* 14||23 */
      New_fils(n,'E',4,'f'); 
      New_fils(n,'I',Num_intern,'f');
      n = n->fin->a->fils ; /* on y descend */
      New_fils(n,'E',2,'f');New_fils(n,'E',3,'f');
      ;break;}
    }
else /* unresolved topology */
  { New_fils(n,'E',2,'f'); New_fils(n,'E',3,'f'); New_fils(n,'E',4,'f'); }
}
/*----------------------------*/
int Vide(Z)
int Z[];
{int x;
/*printf("Vide ? ");*/
for (x=1;x<Etape;x++)
  if (Z[x]) {/*printf("Non\n");*/return(Non);}
/*printf("Oui\n");*/
return(Oui);
}
/*-----------------------------*/
int Hors_de(Sj,sens)
tarete *Sj;int sens;
{/* Le traitement Sup ou Inf est le meme mais doit etre distingue a cause
    de l'orientation de l'arbre et des structures de donnees (bas!=haut..) */
int obj;
listarete *Sl,*Sm,*frere;
tliste *x,*y,*z;
/* Si on a 1 seul elem du cote du parcours, renvoyer ok tout de suite */
if ((sens==Inf) && (!Sj->fils->bas)) return(Oui); /* les Zj restent vides */
if ((sens==Sup) && (!Sj->pere->haut)) return(Oui);
/*------------*/
/* Calcul Z_j */
/*------------*/
if (sens==Inf)
  {
    /* Init le Zj de l'edi Sj par union depuis ses fils */
    for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv)
      for (obj=1;obj<Etape;obj++) 
	if (Sl->a->Z[Inf][obj]) Sj->Z[Inf][obj]=1;
  }
else /* sens = Sup */
  { 
    /* Init le Zj de l'edi Sj par union du Sup du pere et des Inf des freres */
    for (obj=1;obj<Etape;obj++) 
      if (Sj->pere->haut->Z[Sup][obj]) Sj->Z[Sup][obj]=1;
    for (frere=Sj->pere->bas;frere;frere=frere->suiv)
      if (frere->a!=Sj) /* sauf lui-meme */
	for (obj=1;obj<Etape;obj++)
	  if (frere->a->Z[Inf][obj]) Sj->Z[Sup][obj]=1;
  }
/* Retire de Zj les objets de Xj */
for (x=Sj->X[sens];x;x=x->suiv)
  Sj->Z[sens][x->taxon] = 0;

/*printf("Zj =");
for (obj=1;obj<Etape;obj++) if (Sj->Z[sens][obj]) printf(" %d",obj);
printf(".\n");
printf("Xj =");
for (x=Sj->X[sens];x;x=x->suiv) printf(" %d",x->taxon);
printf("\n");*/
/*----------------------*/
/* Examen des I_j^{l,m} */
/*----------------------*/
/* Pour chq couple (Sl,Sm) de sous-arb fils du sous-arb courant Sj */
if (sens==Inf)
  {/*printf("calcul non-trivial\n");*/
  for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv)
   for (Sm=Sl->suiv;Sm;Sm=Sm->suiv)
     { 
       for(x=Sl->a->X[sens];x;x=x->suiv)
         {/*printf("x=%d\n",x->taxon);*/
	 for(y=Sm->a->X[sens];y;y=y->suiv)
	    {/*printf(" y=%d\n",y->taxon);*/
	    for(z=Sj->X[!sens];z;z=z->suiv)
	      {/*printf("  z=%d\n",z->taxon);*/
	      if (!Appartient_Q(x->taxon,y->taxon,z->taxon,Etape)) 
		Sj->Z[sens][z->taxon]=1;
	      }
	    }
	 }
     }
 }
else /* sens = Sup */
  {/*printf("calcul non-trivial : (pere,frere)\n");*/
     /* Pour chq couple (Pere , edi-fils Sl) */
    for (Sl=Sj->pere->bas;Sl;Sl=Sl->suiv)
      if (Sl->a!=Sj)
	for(x=Sj->pere->haut->X[sens];x;x=x->suiv)
	  {/*printf("x=%d\n",x->taxon);*/
	  for(y=Sl->a->X[!sens];y;y=y->suiv)
	    {/*printf(" y=%d\n",y->taxon);*/
	    for(z=Sj->X[!sens];z;z=z->suiv)
	      {/*printf("  z=%d\n",z->taxon);*/
	      if (!Appartient_Q(x->taxon,y->taxon,z->taxon,Etape)) 
		Sj->Z[sens][z->taxon]=1;
	      }
	    }
	  }
   /* Pour chq couple (edi-fils Sl , edi-fils Sm*/
    /*printf("suite : (frere,frere)\n");*/
    for (Sl=Sj->pere->bas;Sl;Sl=Sl->suiv)
      if (Sl->a!=Sj)
	for (Sm=Sl->suiv;Sm;Sm=Sm->suiv)
	  if (Sm->a!=Sj)
	    for(x=Sl->a->X[!sens];x;x=x->suiv)
	      {/*printf("x=%d ",x->taxon);*/
	      for(y=Sm->a->X[!sens];y;y=y->suiv)
		{/*printf(" y=%d ",y->taxon);*/
		for(z=Sj->X[!sens];z;z=z->suiv)
		  {/*printf("  z=%d\n",z->taxon);*/
		  if (!Appartient_Q(x->taxon,y->taxon,z->taxon,Etape)) 
		    Sj->Z[sens][z->taxon]=1;
		  }
		}
	      }
  }
if (Vide(Sj->Z[sens])) return(Oui); else return(Non);
}
/*---------------------------------*/
tliste *Union_Ensembles_X(Xl,Xm)
tliste *Xl,*Xm;
/* SE : Union d'ens disjoints en O(n) car il faut RECOPIER LA LISTE Xm */
{tliste *neo,*x;
for (x=Xm;x;x=x->suiv) 
  {
   neo=(tliste *)malloc(sizeof(tliste));
   if (!neo) Erreur("MEMORY ALLOCATION ERROR (no more space available?)",1);
   neo->taxon=x->taxon;
   neo->suiv=Xl;
   Xl=neo;
  }
return(Xl);
}
/*-----------------------------*/
void Calcule_les_Xj(Sj,sens)
tarete *Sj;int sens;
{listarete *Sl,*frere;
/* Si non trivial remettre Xj a zero */
if ( ((sens==Inf) && (Sj->fils->bas)) || 
     ((sens==Sup) && (Sj->pere->haut)) )
  if (Sj->X[sens])   /* enleve valeur de l'etape precedente */ 
    { /*printf("non trivial ");*/
      /*pr=Sj->X[sens]; for(x=pr->suiv;x;x=x->suiv) free(pr);
      free(pr);*/ 
      Sj->X[sens]=NULL;}

if (sens==Inf) /* faire un parcours Postfixe */
  {
    for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv) 
      { 
	Calcule_les_Xj(Sl->a,sens); /* Parcours et traite les fils */
	/* Calcul le Xj de l'edi Sj par union depuis ses fils */
	Sj->X[Inf]=Union_Ensembles_X(Sj->X[Inf],Sl->a->X[Inf]);
      }
    /*printf("resultat : ");Affiche_Xj(Sj,sens);*/
  }
else /* pr les Sup faire un parcours prefixe */
  {    
   if (Sj->pere!=Taxon1) /* Si racine : Xj dj positionne */
     {
      /* Calcul le Xj depuis l'union du Sup du pere et des Inf des freres */
      Sj->X[Sup]=Union_Ensembles_X(Sj->X[Sup],Sj->pere->haut->X[Sup]);
      for (frere=Sj->pere->bas;frere;frere=frere->suiv)
	{
	 if (frere->a!=Sj) /* sauf lui-meme */
	   Sj->X[Sup]=Union_Ensembles_X(Sj->X[Sup],frere->a->X[Inf]);
	 /*printf("Apres l'union avec le frere Xj = ");Affiche_Xj(Sj,sens);*/
       }
     }
   /* Parcours les fils */
   for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv)
     Calcule_les_Xj(Sl->a,sens);
  }

}
/*-----------------------------*/
void Parcours_Prof(Sj,sens,facon)
tarete *Sj;int sens,facon;
{listarete *Sl;
if (facon==Postfixe)
  {
   for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv) 
     Parcours_Prof(Sl->a,sens,facon);
   Sj->horsde[sens] = Hors_de(Sj,sens);}
else
  {
   Sj->horsde[sens] = Hors_de(Sj,sens);
   for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv) 
     Parcours_Prof(Sl->a,sens,facon);}
}
/*-----------------------------*/
void Enlever_double_non(n)
tnoeud *n;
/* efface les aretes etiquetees par "non" dans les deux sens et
leur noeud inferieur. Rapatrie ses aretes filles en fin des aretes
filles du noeud superieur en O(1) */
{listarete *Sl,*Sm,*prec;
int prem=Non;
Sl=n->bas;prec=n->bas;
while (Sl)
  {
    Enlever_double_non(Sl->a->fils); /* traiter d'abord la sous-arbo */
    if ((Sl->a->horsde[Inf]==Non) && (Sl->a->horsde[Sup]==Non))
      { /* recupere la sous-arbo de l'arete eliminee */
	for (Sm=Sl->a->fils->bas;Sm;Sm=Sm->suiv) Sm->a->pere=n;
	if (Sl->suiv) Sl->a->fils->fin->suiv=Sl->suiv;
	else n->fin=Sl->a->fils->fin; /* il pointe dj sur rien */
	if(prec!=Sl) prec->suiv=Sl->a->fils->bas; 
	else {prem=Oui;prec=n->bas=Sl->a->fils->bas;}
       /* detruit tout ce qui touche au noeud et arete collapsed */
	/* free(Sl->a->fils);
	pr=Sl->a->X[Inf]; if (pr) 
	  { for(x=pr->suiv;x;x=x->suiv) free(pr);  free(pr);}
	pr=Sl->a->X[Sup]; if (pr) 
	  { for(x=pr->suiv;x;x=x->suiv) free(pr);  free(pr);}
	free(Sl->a);free(Sl); */
	/* et ca repart !*/
	if (prem) Sl=prec; else Sl=prec->suiv; 
      }
    else {prec=Sl;Sl=Sl->suiv;} /* ok meme si premier elem */
  }
}
/*-----------------------------*/
void Sur_arete(nd,i)
tnoeud *nd;int i;
{ listarete *Sl;
/*printf("Insere sur arete au dessus de %d\n",nd->taxon);*/
if (nd->bas) /* on est sur un noeud interne */
  {
    /* se cree un nouveau 1er fils */
    New_fils(nd,'I',Num_intern,'d');
    /* lui fait passer toute la sous-arbo d'avant */
    for(Sl=nd->bas->suiv;Sl;Sl=Sl->suiv) Sl->a->pere=nd->bas->a->fils;
    nd->bas->a->fils->bas=nd->bas->suiv;
    nd->bas->a->fils->fin=nd->fin;
    nd->bas->suiv=NULL; nd->fin = nd->bas;
  }
else /* on est sur un noeud externe */
  { /* on transfert le taxon en 1 neo noeud fils */
    New_fils(nd,'E',nd->taxon,'d');
    nd->taxon=Num_intern;Num_intern++;
  }
/* rajoute en 2eme fils le neo taxon */
New_fils(nd,'E',i,'f');    
}
/*-----------------------------*/
void Sur_noeud(nd,i)
tnoeud *nd;int i;
{/*printf("Insere sur noeud %d\n",nd->taxon);*/
New_fils(nd,'E',i,'f');
}
/*-----------------------------*/
void Insere_tax_arbre(n,i,ok)
tnoeud *n;int i,*ok;
/* deduit l'endroit ou greffer le taxon i depuis les etiquettes des aretes
   de l'arbre. `ok' sert a arreter la descente en prof (on depile juste) */
{listarete *Sl;
for (Sl=n->bas;(Sl) && (*ok==Non);Sl=Sl->suiv)
  {
    if (Sl->a->horsde[Sup]==Oui)
      if (Sl->a->horsde[Inf]==Oui) {Sur_arete(Sl->a->fils,i); *ok=Oui;}
      else 
	if (Sl->a->fils->bas) Insere_tax_arbre(Sl->a->fils,i,ok);
  }
 if (!(*ok)) /* tout indique le noeud courant */ 
   {Sur_noeud(n,i);*ok=Oui;}
}
/*-----------------------------*/
void Raz_Zj(Sj)
tarete *Sj;
{int obj;listarete *Sl;
for (obj=1;obj<Etape;obj++) {Sj->Z[Inf][obj]=0;Sj->Z[Sup][obj]=0;}
for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv) Raz_Zj(Sl->a);
}
/*-----------------------------*/
void Inclus_neo_tax(i)
int i;
{int ok=Non;
/*printf("Calcule les Xj des edi Inf\n");*/
Calcule_les_Xj(Taxon1->bas->a,Inf);
/*printf("calcule les Xj des edi Sup (forcement apres les Inf)\n");*/
Calcule_les_Xj(Taxon1->bas->a,Sup);

/* necessaire ? */
/* Si oui peut etre combine avec Calcul des Xj ci-dessus */
/* Si on l'enleve on a pas les memes resultats dans certains cas !!! */ 
Raz_Zj(Taxon1->bas->a);

/*printf("parcours 1 en prof pr les etiquettes\n");*/
Parcours_Prof(Taxon1->bas->a,Inf,Postfixe);
/*printf("parcours 2 en prof pr les etiquettes\n");*/
Parcours_Prof(Taxon1->bas->a,Sup,Prefixe);
/*printf("Apres avoir teste insertion de l'objet %d\n",i);
Affiche_Arbre(Taxon1);*/
Enlever_double_non(Taxon1);
/*printf("Apres avoir enleve les aretes non-supportees :\n");
Affiche_Arbre(Taxon1);*/
Insere_tax_arbre(Taxon1,i,&ok);
/*printf("Apres avoir insere le nouvel objet:\n");
Affiche_Arbre(Taxon1);*/
}
/*--------------------------*/
void State_Bipartitions(Sj)
tarete *Sj;
{listarete *Sl;tliste *x;
int t,ind; 
Nb_bips++;
/* clean from last use */
for (t=0;t<Nb_taxa;t++) up_part[t]=down_part[t]=0; 
/* Fill current bip from taxa above and below current edge */
for (x=Sj->X[Inf],ind=0;x;x=x->suiv,ind++)  
  down_part[x->taxon-1]=1; /* -1 because taxa numbered from 1 and not from 0 */
for (x=Sj->X[Sup],ind=0;x;x=x->suiv,ind++)  
  up_part[x->taxon-1]=1;   /* -1 because taxa numbered from 1 and not from 0 */
/* save to file */
 for(t=0;t<Nb_taxa;t++)  if (down_part[t]) fprintf(fb,"%.2d ",t+1); /* +1 because ... */
 fprintf(fb,"|");
 for(t=0;t<Nb_taxa;t++)  if (up_part[t]) fprintf(fb," %.2d",t+1); /* +1 because ... */
 fprintf(fb,"\n");
/* recursively process children */
for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv) 
  State_Bipartitions(Sl->a);
}

/*-----------------------------*/
void Save_Bipartitions()
{
int t;
/* Allocate space for the 2 parts of a "current" bipartition */
 if (! (up_part=(short *)malloc(Nb_taxa*sizeof(short))) ) /* 0/1 infos */
  Erreur("MEMORY ALLOCATION ERROR (no more space available?)",1);
 if (! (down_part=(short *)malloc(Nb_taxa*sizeof(short))) ) /* 0/1 infos */
  Erreur("MEMORY ALLOCATION ERROR (no more space available?)",1);
 /* Save taxa name % nb correspondance */
fb=fopen("bipfile","w");
fprintf(fb,"   %d\n",Nb_taxa);
for (t=1;t<=Nb_taxa;t++)
  fprintf(fb,"%.2d %s\n",t,name[t]);
fprintf(fb,"\n");
printf("Saving bipartitions (edges) of the tree");
/* Compute trivial (for addquart) and non-trivial bips induced by the tree */
/* Recompute Xj's to account for last species added to the tree */
Calcule_les_Xj(Taxon1->bas->a,Inf);
Calcule_les_Xj(Taxon1->bas->a,Sup);
/* Recursive search of the tree, each edge induces a bipartition */
State_Bipartitions(Taxon1->bas->a);
fclose(fb);
printf(" in bipfile \n");
printf("Resolution of the tree = %.2lf%% (# internal edges inferred / max # internal edge)\n",(double)(Nb_bips-Nb_taxa)*100.0/(Nb_taxa-3));
}
/*-----------------------------*/
void Save_Crossed_quartets(edge)
tarete *edge;
/* SE : The crossed quartets are those resolved by the inferred tree 
   (in the same way as in the input quartet set) */
/* SI : s'inspire de Calcule_les_Xj(..,Inf) */
{
listarete *childedge,*c1,*c2;
tliste *tl,*tr,*tu1,*tu2;
int x,y,z,t;
/* Traverse and process child edges */
for (childedge=edge->fils->bas;childedge;childedge=childedge->suiv) 
     Save_Crossed_quartets(childedge->a); 

/* These intricated loops are in O(n^4) (surprising isn't it?!!!).
This results from the fact that each quartet is a crossed quart for at most
two edges (and more often of only one) :
e.g.

  |     |
 / \   / \
 2 3   6 14        */
/* Compute crossed quartets associated to the edge */
if ((edge->X[Sup]->suiv) /* if not on trivial edge set up as artificial root */
     && (edge->fils->bas)) /* and not on a pendant trivial edge */ 
   {
     /* For each pair (c1,c2) of child edges (any non-trivial edge has at least 2 children) */
     for (c1=edge->fils->bas;c1->suiv;c1=c1->suiv)
       for (c2=c1->suiv;c2;c2=c2->suiv)
	 /* For each taxa t1 below c1 and t2 below c2 */
	 for (tl=c1->a->X[Inf];tl;tl=tl->suiv)
	   for (tr=c2->a->X[Inf];tr;tr=tr->suiv)
	     /* For each pair of taxa above the current edge */
	     for (tu1=edge->X[Sup];tu1->suiv;tu1=tu1->suiv)
	       for (tu2=tu1->suiv;tu2;tu2=tu2->suiv)
		 {
		   x=tl->taxon;y=tr->taxon;z=tu1->taxon;t=tu2->taxon;
		   /* order species preserving quartet relationship */
		   Order4tree(&x,&y,&z,&t);
		   /*printf("%d %d || %d %d\n",x,y,z,t);*/
		   /* Set unresolved in struct Quartets */ 
		   if (SetQuartet(x,y,z,t,0))
		     { /* Sauver ds quartfile.res */
		     fprintf(fr,"%.2d %.2d || %.2d %.2d\n",x,y,z,t);
		     NbQuartRes++;
		     }
		 }
   }
}
/*-----------------------------*/
void Sauve_Quads()
{
/* met les quadruplets resolus dans "quartfile.res" 
et les autres dans "quartfile.left" */
int t1,t2,t3,t4;
double cn4;
int list[4];
/* Output quartets of Q resolved by the inferred tree and remove them from Q */
fr=fopen("quartfile.res","w");fprintf(fr,"   %d\n",Nb_taxa);
for (t1=1;t1<=Nb_taxa;t1++) fprintf(fr,"%.2d %s\n",t1,name[t1]);
fprintf(fr,"\n");
Save_Crossed_quartets(Taxon1->bas->a);
fclose(fr);

/* Output quartets remaining in Q */
fl=fopen("quartfile.left","w");fprintf(fl,"   %d\n",Nb_taxa);
for (t1=1;t1<=Nb_taxa;t1++) fprintf(fl,"%.2d %s\n",t1,name[t1]);
fprintf(fl,"\n");

for (t1=1;t1<=Nb_taxa-3;t1++)
  for(t2=t1+1;t2<=Nb_taxa-2;t2++)
    for (t3=t2+1;t3<=Nb_taxa-1;t3++)
      for(t4=t3+1;t4<=Nb_taxa;t4++)
	if (GetQuartet(t1,t2,t3,t4,list))
           { /* list[0] is x */
	     if (list[1]==t2) /* t1t2||t3t4 */
	       fprintf(fl,"%.2d %.2d || %.2d %.2d\n",t1,t2,t3,t4);
	     else if (list[1]==t3)  /* t1t3||t2t4 */
	       fprintf(fl,"%.2d %.2d || %.2d %.2d\n",t1,t3,t2,t4);
	     else  /* t1t4||t2t3 */
	       fprintf(fl,"%.2d %.2d || %.2d %.2d\n",t1,t4,t2,t3);
	   }      
fclose(fl);
/* A ARRANGER :  */
printf("%ld quartets in Q* (output in file : quartfile.res)\n",NbQuartRes);
printf("%ld quartets left in Q (output in file : quartfile.left)\n",NbQuartRead-NbQuartRes);
printf("Treelikeness of the input quartet set = %.2lf%% (# quartets in the tree / # input quartets)\n",(double)(NbQuartRes)*100.0/(double)NbQuartRead); 
 cn4=(double)(((long)Nb_taxa)*(Nb_taxa-1)*(Nb_taxa-2)*(Nb_taxa-3))/(4*3*2);
  printf("Treelikeness wrt fully resolved tree = %.2lf%% (# quartets in the tree / # possible quartets)\n",(double)(NbQuartRes)*100.0/cn4);
}
/*---------------------------------------------------------------*/
/*---------------------------------------------------------------*/        

int main(argc, argv)
int argc;char *argv[];
{extern char *optarg;
char c;
/* Reads set Q of quartets from file (and init Nb_taxa) */
Lit_quadruplets();
if (!NbQuartRead) printf("ERROR ? quartfile is EMPTY !!!!\n");
/* T_4 = resolution du quadruplet (1,2,3,4) */
Etape=4;Num_intern=Max_nb_taxa+1;
Init_Arbre();
/*Affiche_Arbre(Taxon1);*/
printf("\nIncorporating taxa in the tree :\n....");
/* Insere une nouvelle espce a chaque etape */
for (Etape=5;Etape<=Nb_taxa;Etape++)
    {Inclus_neo_tax(Etape);printf(".");fflush(stdout);}
printf("\n");
/* Sauve les bipartitions B*  */
Save_Bipartitions();
/*Affiche_Arbre(Taxon1);getchar();*/
Sauve_Quads(); 
}

